# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

lappend auto_path [file join [pwd] lib]

package require Tk 8.6
package require Ttk 8.6
package require cmdbar
package require config 
package require macro

package require audio 1.5
package require TkEsweepXYPlot
package require overlay 1.3

# include local packages
source [file join [pwd] lib config.tcl]
source [file join [pwd] lib gui.tcl]
source [file join [pwd] lib dsp.tcl]
source [file join [pwd] lib math.tcl]
source [file join [pwd] lib file.tcl]

####################
# global variables #
####################

# ID of master curve
set ::master 1

trace add variable ::master write [list apply {{v1 v2 op} {
    upvar $v1 master
    DEBUGPRINT "making ID #$master master curve"
    ::overlay::removeTag all master
    ::overlay::addTag $master master -background orange
    :cursor trace 1 #$master
    set ::diff(Master) $master
    # we need to update the harmonic distortion
    #updateHD
    #if {[::TkEsweepXYPlot::isVisible HD]} {::TkEsweepXYPlot::plot HD}
  }}]

# arrange for cursor difference readout
# this is done with a hidden trace in the display, which is also not listed as an overlay
array set ::diff {
  Master  {}
  Slave   {}
  ID 0
}

# the updateDiff cmd is in gui.tcl!
trace add variable ::diff(Master) write updateDiff
trace add variable ::diff(Slave) write updateDiff

######################
# Overlay management #
######################

proc initOverlays {w} {
  ::overlay::addProperty Color      {}        0  {Display color}
  ::overlay::addProperty Name       {}        200 {Descriptive name}
  ::overlay::addProperty IR         {}        0  {Measured impulse response}
  ::overlay::addProperty GatedIR    {}        0  {Impulse response windowed with the current gate}
  ::overlay::addProperty FR         {}        0  {Measured frequency response}
  ::overlay::addProperty SigType    {}        0  {Measurement signal type}
  ::overlay::addProperty SweepRate  {}        0  {Sweep rate, only useful when SigType is lin- or logsweep}
  ::overlay::addProperty Gate       {0 end}   0  {Impulse response gating window position}
  ::overlay::addProperty Window     {}        0  {Impulse response gating window type}
  ::overlay::addProperty Show       {yes}     0  {Show or hide this overlay (boolean)}
  ::overlay::addProperty Mode       {single}  0  {Processing mode}
  ::overlay::addProperty OutLvl     {2.83}    0  {Output level (in V)}
  ::overlay::addProperty Mic        {}        0  {Microphone compensation, short form, saved}
  ::overlay::addProperty MicExp     {}        0  {Microphone compensation, expanded form, not saved}

  set win [::overlay::init $w]
  set ::overlay::selectFunc [list updateSelection {{FR 4 2} {IR 2 0} {BD 2 0}}]

  pack $win -side right -expand yes -fill both
}

proc updateSelection {plots} {
  set allids [lsort -increasing -integer [::overlay::ids]]
  foreach def $plots {
    lassign $def P width default
    foreach id [::overlay::select get] {
      catch {::TkEsweepXYPlot::configTrace $P $id -width $width}
      set i [lsearch -sorted -integer $allids $id]
      set allids [lreplace $allids $i $i]
    }
    foreach id $allids {
      catch {::TkEsweepXYPlot::configTrace $P $id -width $default}
    }
    catch {::TkEsweepXYPlot::plot $P}
  }
}

# This is the main updating procedure for the plots
# The first time it gets called, it logs the ID, the property, and the operation, then starts a timer. 
# When this timer expires, it updates the plot. 
# If the function gets called again _before_ the timer expires, it appends the new parameters to its log, 
# and resets the timer. This way, we can avoid multiple, unnecessary redraws of the plot
proc frUpdateFunc {args} {
  # get unique names for timer variables
  set waiter [uniqueName]
  set timer [uniqueName]
  while 1 {
    set log [list]
    set $waiter 1
    lappend log {*}[yieldto return -level 0]
    # next call enters the logger
    while 1 {
      updateTimer $timer $waiter

      lappend log {*}[yieldto return -level 0]
      if {![set $waiter]} {break}
    }

    foreach {nu id overlays prop op} $log {
      switch -- $op {
        unset {
          catch {::TkEsweepXYPlot::removeTrace FR $id}
          catch {::TkEsweepXYPlot::removeTrace IR $id}
          catch {::TkEsweepXYPlot::removeTrace BD $id}
        }
        write -
        default {
          if {[info exists flags($id)] && [lsearch $flags($id) $prop] >= 0} {
            DEBUGPRINT "Property $prop of #$id is up to date"
            continue
          } 
          switch -- $prop {
            Show {
              setVisibility $id
              lappend flags($id) Show
            }
            Color {
              set color [getProp $id Color]
              catch {::TkEsweepXYPlot::configTrace FR $id -color $color}
              catch {::TkEsweepXYPlot::configTrace IR $id -color $color}
              catch {::TkEsweepXYPlot::configTrace BD $id -color $color}
              lappend flags($id) Color
            }
            Name {
              set name [getProp $id Name]
              catch {::TkEsweepXYPlot::configTrace FR $id -name $name}
              catch {::TkEsweepXYPlot::configTrace IR $id -name $name}
              catch {::TkEsweepXYPlot::configTrace BD $id -name $name}
              # the HD window only works on the master curve
              if {$id == $::master} {updateHD $id}
              lappend flags($id) Name
            }
            MicExp -
            Gate -
            Window -
            IR {
              # update gated IR and FR for standard plot updates
              updateGatedIR $id
              updateFR $id
              # updates other plots
              updateIR $id
              updateBD $id
              # the HD window only works on the master curve
              if {$id == $::master} {updateHD $id}
              lappend flags($id) MicExp Gate Window IR
            }
            default {
            }
          }
        }
      }
    }
    array unset flags
    ::TkEsweepXYPlot::plot FR
    if {[::TkEsweepXYPlot::isVisible IR]} {::TkEsweepXYPlot::plot IR}
    if {[::TkEsweepXYPlot::isVisible HD]} {::TkEsweepXYPlot::plot HD}
    if {[::TkEsweepXYPlot::isVisible BD]} {::TkEsweepXYPlot::plot BD}

    DEBUGPRINT "Done"
  }
}
# register the coroutine context as the update function
set ::overlay::updateProps [list Name IR Color Mode Gate Window MicExp Show]
set ::overlay::updateFunc [list [spawn frUpdateFunc] {_}]

##################
# Macro handling #
##################

if {0} {
proc load_macro {{name {}}} {
  if {$name eq {}} {
    foreach fname [glob -nocomplain -directory [file join [pwd] macros] *.frm] {
      ::macro::import $fname
    }
  } else {
    set fname [file join [pwd] macros $name.frm]
    ::macro::import $fname
  }
}
}

proc init_macros {} {
  load_macro
  ::macro::addAlias :w :l :set :measure :delete :expr :show :hide :q :export :import :rename :modify :message :copy :paste :query :list :clip :select :dirac :get :wait :filter
  if {$::Debug} {
    ::macro::addAlias puts
  }
}

proc :measure {{mic default} {insertMode replace}} {
  global master
  set FS [setup cget Audio,Samplerate]
  set ref [esweep::create -type wave -samplerate $FS -size [expr {int(0.5+$FS*[setup cget Signal,Length]/1000.0)}]]
  switch -- [setup cget Measure,Mode] {
    external {
      if {[catch {set ref [esweep::load -filename [file join [pwd] [setup cget Signal,External,Reference]]]}]} {
        ::cmdbar::writeToCmdBar \
          "Could not read reference file [setup cget Signal,External,Reference]" -bg red
          return
      }
      setup configure Audio,Samplerate [esweep::samplerate -obj $ref]
    }
    single -
    dual {
      set sweeprate [esweep::generate [setup cget Signal,Type] -obj ref \
                        -locut [lindex [setup cget Signal,Bandwidth] 0] \
                        -hicut [lindex [setup cget Signal,Bandwidth] 1] \
                        -spectrum [setup cget Signal,Spectrum]]
    }
  }
  if {[setup cget Signal,Type] ne {logsweep}} {
    set sweeprate -1
  }

  # compute output level
  set ch_out [setup cget Output,Channel]
  if {[catch {setup cget Output,$ch_out,Gain}] || [catch {setup cget Output,$ch_out,Sensitivity}]} {
    set ch Global
    ::cmdbar::writeToCmdBar \
      "Output channel $ch_out not fully specified, using global output parameters" -bg yellow
  }
  set outsense [setup cget Output,$ch_out,Sensitivity]
  set outgain [setup cget Output,$ch_out,Gain]
  set lvl [expr {1.414*[setup cget Output,Level] / ([setup cget Output,$ch_out,Gain] * [setup cget Output,$ch_out,Sensitivity])}]
  if {$lvl > [setup cget Output,MaxFS]} {
    set lvl [setup cget Output,MaxFS]
    setup configure Output,Level [format "%.3f" [expr \
      {$lvl*[setup cget Output,$ch_out,Gain]*[setup cget Output,$ch_out,Sensitivity]/1.414}]]
    ::cmdbar::writeToCmdBar \
      "Output level exceeded maximum allowed output. Adjusted to [setup cget Output,Level] V" -bg yellow
  }
  set lvl [expr {$lvl/[esweep::max -obj $ref]}]
  esweep::expr ref * $lvl

  # calculate scaling
  set ch_ref [setup cget Input,Reference]
  if {[catch {setup cget Input,$ch_ref,Gain}] || [catch {setup cget Input,$ch_ref,Sensitivity}]} {
    set ch_ref Global
    ::cmdbar::writeToCmdBar \
      "Reference channel $ch_ref not fully specified, using global input parameters" -bg yellow
  }
  set ch_in [setup cget Input,Channel]
  if {[catch {setup cget Input,$ch_in,Gain}] || [catch {setup cget Input,$ch_in,Sensitivity}]} {
    set ch_in Global
    ::cmdbar::writeToCmdBar \
      "Input channel $ch_in not fully specified, using global input parameters" -bg yellow
  }
  set insense [setup cget Input,$ch_in,Sensitivity]
  set ingain [setup cget Input,$ch_in,Gain]
  set refsense [setup cget Input,$ch_ref,Sensitivity]
  set refgain [setup cget Input,$ch_ref,Gain]
  if {$mic eq {default}} {set mic [setup cget Mic,Default]}
  set micsense [setup cget Mic,$mic,Sensitivity]
  set micgain [setup cget Mic,$mic,Gain]
  switch -- [setup cget Measure,Mode] {
    single {
      set sc [expr {1000.0*$insense/($micsense*$outsense*$outgain*$ingain*$micgain)}]
    }
    dual {
      set sc [expr {1000.0*$refgain*$insense/($micsense*$refsense*$ingain*$micgain)}]
    }
    external {
      set sc [expr {1000.0*$insense/($micsense*$ingain*$micgain)}]
    }
    default {
        return -code error "Unknown processing mode [setup cget Measure,Mode]" 
    }
  }

  set dist [expr {[setup cget Mic,$mic,Distance] < 0 ? 1.0 : [setup cget Mic,$mic,Distance]}]
  set record_length [expr \
    {[setup cget Signal,Length]+[setup cget Audio,Latency]+(1000.0*$dist)/[setup cget Physics,C]}]

  switch -- [setup cget Measure,Mode] {
    external {
        set b [::cmdbar::deleteBindings]
        bind . <space> [list ::audio::sconfigure Trigger 2]
        lassign [::audio::measure external -ref ref -trigger [list [setup cget Processing,External,Trigger,Type] [expr {pow(10, [setup cget Processing,External,Trigger,Level]/20.0)}]] -wait 1000] ir
        bind . <space> {}
        ::cmdbar::restoreBindings $b
    }
    single -
    dual {
      set ir [::audio::measure [setup cget Measure,Mode] \
                              -ref ref \
                              -input [list [setup cget Input,Channel] [setup cget Input,Reference]] \
                              -output [setup cget Output,Channel] \
                              -record $record_length]
    }
    default {
        return -code error "Unknown processing mode [setup cget Measure,Mode]"
    }
  }

  if {[catch {esweep::size -obj $ir}]} {
    ::cmdbar::writeToCmdBar {Measurement cancelled} -bg yellow
    return
  }
  if {[setup cget Processing,Gate,Mode] eq {auto}} {
    # Place start of gate automatically. In single channel mode, use 10 ms in front of IR maximum.
    # In dual channel mode, set to 0 ms + microphone distance correction.
    switch -- [setup cget Measure,Mode] {
      external -
      single {
        set max [esweep::max -obj $ir]
        set min [esweep::min -obj $ir]
        set start [expr {1000.0*($max > $min ? [esweep::maxPos -obj $ir] : [esweep::minPos -obj $ir])/$FS-10}]
        if {$start < 0} {set start 0.0}
      }
      dual {
        set dist [expr {[setup cget Mic,$mic,Distance] < 0 ? 0.0 : [setup cget Mic,$mic,Distance]}]
        set start [expr {1000.0*$dist/[setup cget Physics,C]}]
      }
      default {
          return -code error "Unknown processing mode [setup cget Measure,Mode]"
      }
    }
  } else {
    set start [lindex [setup cget Processing,Gate] 0]
  }

  esweep::expr ir * $sc
  set props [list Name Master Color red \
                SigType [setup cget Signal,Type] \
                SweepRate $sweeprate \
                Gate [list $start [expr {[lindex [setup cget Processing,Gate] end]+$start}]] \
                Window [list [setup cget Processing,Gate,Window,Type] [setup cget Processing,Gate,Window,Width]] \
                Mode [setup cget Measure,Mode] \
                OutLvl [setup cget Output,Level] \
                IR $ir \
                Mic [setup cget Mic,$mic,Compensation] \
                MicExp [expandMicCompensation [setup cget Mic,$mic,Compensation]] \
                Show {yes}]

  if {$insertMode eq {add} || ![::overlay::exists $::master]} {
    set ::master [::overlay::add $props]
    # add the impulse response to force a display update
    ::overlay::modify $::master [list IR $ir]
  } elseif {$insertMode eq {replace}} {
    # set the ::master here again to trigger some updates
    set ::master [::overlay::modify $::master $props]
  } else {
    return -code error "Unknown insert mode $insertMode"
  }
  ::overlay::select set $::master
  # copy master to buffer
  ::overlay::copy
  update

  return #$::master
}

# we need our own version of :paste to copy the impulse response
proc :paste {} {
  foreach id [::overlay::paste] {
    set props [list Name "Copy of [getProp $id Name]" IR [esweep::clone -src [getProp $id IR]]]
    ::overlay::modify $id $props
    lappend IDs #$id
  }
  # join the list with spaces, workaround because the first ID may have been added as a sublist
  return [join $IDs { }]
}

proc :delete {ID} {
  if {$ID ne {all}} {
    if {![regexp {(^#)([[:digit:]]+$)} $ID -> hash id]} {
        return -code error "Parameter not an ID: $ID"
    }
  } else {
    set id $ID
  }
  ::overlay::delete $id
  return
}

proc :get {ID p1 p2} {
  if {[catch {isID $ID} id]} {
    return -code error $id
  }
  switch -nocase -glob -- $p1 {
    Ph* {
      # get phase at frequency p2
      set fr [getProp $id FR]
      # mapping
      set f0 [lindex $fr 0]
      set f1 [lindex $fr end-2]
      if {$p2 < $f0 || $p2 > $f1} {
        return -code error "Frequency $p2 Hz out of bounds"
      }
      set n [expr {[llength $fr]/3}]
      set i [expr {3*int(log10($p2/$f0)*$n/log10($f1/$f0))}]
      set res [interpolate [lindex $fr $i] [lindex $fr $i+2] [lindex $fr $i+3] [lindex $fr $i+5]]
    }
    Mag* -
    default {
      # get magnitude at frequency p2
      set fr [getProp $id FR]
      # mapping
      set f0 [lindex $fr 0]
      set f1 [lindex $fr end-2]
      if {$p2 < $f0 || $p2 > $f1} {
        return -code error "Frequency $p2 Hz out of bounds"
      }
      set n [expr {[llength $fr]/3}]
      set i [expr {3*int(log10($p2/$f0)*$n/log10($f1/$f0))}]
      set res [interpolate [lindex $fr $i] [lindex $fr $i+1] [lindex $fr $i+3] [lindex $fr $i+4] $p2]
    }
  }
  return $res
}


###########
# Startup #
###########
default_config
appWindows
bindings .
load_config
init_macros
readMicCompensation all
