# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide overlay 1.3

package require Tk
package require Ttk

namespace eval ::overlay {
  variable properties
  array set properties {}

  variable overlays
  array set overlays {}

  variable buffer
  array set buffer {}

  variable id
  set id 0

  variable win
  set win {}

  # call this function every time...
  variable updateFunc
  set updateFunc {} 
  # ... a property listed in updateProps gets modified
  variable updateProps
  set updateProps {}

  # this function will be called when the selection changes
  variable selectFunc
  set selectFunc {}


  namespace ensemble create -command overlay -subcommands [list add modify delete get copy yank]
}

proc ::overlay::init {w} {
  variable win
  variable properties
  variable selectFunc
  if {$w eq {.}} {set w {}}
  set win $w.overlay
  set cols [list]
  foreach prop [array names properties] {
    if {[lindex $properties($prop) 1] > 0} {
      lappend cols $prop
    }
  }
  ttk::treeview $win -columns $cols
  $win heading #0 -text ID
  $win column #0 -width 50
  foreach col $cols {
    $win heading $col -text $col
    $win column $col -width [lindex $properties($col) 1]
  }

  $win tag configure disabled -background darkgrey

  trace add variable selectFunc write [list apply {{win v1 v2 op} {
      upvar $v1 v
      bind $win <<TreeviewSelect>> $v
    }} $win]

  return $win
}

proc ::overlay::addProperty {name default {dispWidth 50} {descr {}}} {
  variable properties
  if {[info exists properties($name)]} {
    return -code error "Property \"$name\" exists."
  }
  set properties($name) [list $default $dispWidth $descr]
  return $properties($name)
}

proc ::overlay::modifyProperty {name {default {}} {dispWidth {}} {descr {}}} {
  variable properties
  if {![info exists properties($name)]} {
    return -code error "Property \"$name\" does not exist."
  }
  if {$default eq {}} {set default [lindex $properties($name) 0]}
  if {$dispWidth eq {}} {set dispWidth [lindex $properties($name) 1]}
  if {$descr eq {}} {set descr [lindex $properties($name) 2]}
  return [set properties($name) [list $default $dispWidth $descr]]
}

proc ::overlay::getProperty {{name {}}} {
  variable properties
  if {$name eq {}} {
    # return all property names
    return [array names properties]
  }
  if {![info exists properties($name)]} {
    return -code error "Property \"$name\" does not exist."
  }
  return $properties($name)
}

proc ::overlay::deleteProperty {name} {
  variable properties
  if {$name eq {}} {
    # delete array
    return [array set properties {}]
  }
  if {![info exists properties($name)]} {
    return -code error "Property \"$name\" does not exist."
  }
  return [array unset properties $name]
}

proc ::overlay::win {} {
  variable win
  return $win
}

proc ::overlay::add {{propList {}}} {
  variable properties
  variable overlays
  variable id
  variable win

  incr id
  foreach prop [array names properties] {
    lassign $properties($prop) def
    set overlays($id,$prop) $def
  }
  foreach {prop value} $propList {
    if {![info exists overlays($id,$prop)]} {
      return -code error "Property \"$prop\" does not exist."
    }
    set overlays($id,$prop) $value
  }

  # extract displayable values
  # note the workaround for the faulty configure command result
  set valList {}
  foreach col [lindex [$win configure -columns] end] {
    if {[set i [lsearch $propList $col]] >= 0 && $i % 2 == 0} {
      lappend valList [lindex $propList $i+1]
    } else {
      lappend valList {}
    }
  }

  # insert overlay
  set item [$win insert {} end -id $id -text $id -values $valList]
  # if there's a Color property, use it as the label color
  if {[info exists overlays($id,Color)]} {
    set im [image create photo -width 10 -height 10]
    $im put $overlays($id,Color) -to 0 0 10 10
    $win item $item -image $im
    trace add variable overlays($id,Color) unset [list apply {{im v1 v2 op} {
        image delete $im
        update idletasks
      }} $im]
    trace add variable overlays($id,Color) write [list apply {{im v1 v2 op} {
        upvar $v1 overlays
        $im put $overlays($v2) -to 0 0 10 10
        update idletasks
      }} $im]
  }
  trace add variable overlays([lindex [array names overlays $id,*] 0]) unset [list apply {{win item v1 v2 op} {
      $win delete $item
    }} $win $item]

  installUpdFunc $id

  $win selection set $item
  return $id
}

proc ::overlay::get {ID {prop {}}} {
  variable overlays
  variable properties
  variable id

  set IDs [list]
  if {$ID eq {all}} {
    for {set i 1} {$i <= $id} {incr i} {
      if {[llength [array names overlays $i,*]] > 0} {
        lappend IDs $i
      }
    }
  } else {
    if {[llength [array names overlays $ID,*]] <= 0} {
      return -code error "Overlay \"$ID\" does not exist."
    }
    lappend IDs $ID
  }

  if {$prop eq {}} {
    set prop *
  } elseif {![info exists overlays($ID,$prop)]} {
    return -code error "Property \"$prop\" does not exist."
  }
  set ret [list]
  foreach ID $IDs {
    lappend ret $ID
    set pv [list]
    foreach {p v} [array get overlays $ID,$prop] {
      lappend pv [join [lrange [split $p ,] 1 end] ,] $v
    }
    lappend ret $pv
  }
  return $ret
}

proc ::overlay::ids {} {
  variable overlays
  variable id
  set IDs [list]
  for {set i 1} {$i <= $id} {incr i} {
    if {[llength [array names overlays $i,*]] > 0} {
      lappend IDs $i
    }
  }
  return $IDs
}

proc ::overlay::exists {ID} {
  variable overlays
  variable properties

  foreach prop [array names properties] {break}

  return [info exists overlays($ID,$prop)]
}

proc ::overlay::modify {ID propList} {
  variable overlays
  variable win
  foreach {prop value} $propList {
    if {![info exists overlays($ID,$prop)]} {
      return -code error "Overlay \"$ID\" or property \"$prop\" does not exist."
    }
    #DEBUGPRINT "$ID,$prop: [info level 1]"
    set overlays($ID,$prop) $value
    foreach col [lrange [$win configure -columns] 4 end] {
      set values [$win item $ID -values]
      lset values [lsearch [lrange [$win configure -columns] 4 end] $col] $overlays($ID,$col)
      $win item $ID -values $values
    }
  }
  return $ID
}

proc ::overlay::addTag {ID tag args} {
  variable win
  if {$ID eq {all}} {
    set ID [::overlay::ids]
  }
  foreach id $ID {
    $win tag add $tag $id
  }
  eval $win tag configure $tag $args
}

proc ::overlay::removeTag {ID tag} {
  variable win
  if {$ID eq {all}} {
    set ID [::overlay::ids]
  }
  foreach id $ID {
    $win tag remove $tag $id
  }
}

proc ::overlay::disable {ID} {
  variable overlays
  variable win
  if {$ID eq {}} {return}
  $win tag add disabled $ID
}

proc ::overlay::enable {ID} {
  variable overlays
  variable win
  if {$ID eq {}} {return}
  $win tag remove disabled $ID
}

proc ::overlay::delete {{ID {}}} {
  variable overlays
  variable id
  if {$ID eq {all}} {
    set ID *
    set id 0
  }
  if {$ID ne {}} {array unset overlays $ID,*}
}

proc ::overlay::select {op {ID {}}} {
  variable overlays
  variable win
  switch -- $op {
    get {
      return [$win selection]
    }
    prev {-}
    next {
      set ID [lindex [$win selection] 0]
      if {[set n [$win $op $ID]] ne {}} {
        $win selection set $n
      }
    }
    end {
      while {[set n [$win next [lindex [$win selection] 0]]] ne {}} {
        $win selection set $n
      }
    }
    set {-}
    add {-}
    remove {-}
    toggle {
      return [$win selection $op $ID]
    }
    default {
      return -code error "Unknown selection operator \"$op\""
    }
  }
}

# copy currently selected overlay
proc ::overlay::copy {{ID {}}} {
  variable overlays
  variable buffer
  variable win

  # clear buffer
  array unset buffer
  if {$ID eq {}} {set ID [$win selection]}
  foreach id $ID {
    foreach tok [array names overlays $id,*] {
      set buffer($tok) $overlays($tok)
    }
  }
  return $ID
}

proc ::overlay::yank {} {
  variable overlays
  variable buffer
  variable win

  # clear buffer
  array unset buffer
  foreach ID [set sel [$win selection]] {
    foreach tok [array names overlays $ID,*] {
      set buffer($tok) $overlays($tok)
    }
    array unset overlays $ID,*
  }
  return $sel
}

proc ::overlay::paste {} {
  variable overlays
  variable buffer
  variable win
  variable id

  set sel [$win index [lindex [$win selection] end]]

  set IDs [list]
  foreach tok [array names buffer *,Name] {
    lappend IDs [lindex [split $tok ,] 0]
  }

  if {[llength $IDs] <= 0} {
    # nothing to paste
    return {}
  }

  # extract displayable values
  set propList [list]
  foreach ID $IDs {
    foreach token [array names buffer $ID,*] {
      regexp -- {^[[:digit:]]+,([[:print:]]+)$} $token -> prop
      lappend propList $prop
      #lappend propList [lrange [split $token ,] 1 end]
    }
    # note the workaround for the faulty configure command result
    foreach col [lrange [$win configure -columns] 4 end] {
      if {[set i [lsearch $propList $col]] >= 0} {
        lappend valList $buffer($ID,$col)
      } else {
        lappend valList {}
      }
    }
    set propList [list]

    # insert the item after the currently selected with a new ID
    incr id
    lappend ret $id
    # copy buffer to overlays
    foreach token [array names buffer $ID,*] {
      regexp -- {^[[:digit:]]+,([[:print:]]+)$} $token -> prop
      #set prop [join [lrange [split $tok ,] 1 end]]
      set overlays($id,$prop) $buffer($token)
    }
    # insert overlay
    set item [$win insert {} [incr sel] -id $id -text $id -values $valList]
    # if there's a Color property, use it as the label color
    if {[info exists overlays($id,Color)]} {
      set im [image create photo -width 10 -height 10]
      $im put $overlays($id,Color) -to 0 0 10 10
      $win item $item -image $im
      trace add variable overlays($id,Color) unset [list apply {{im v1 v2 op} {
          image delete $im
          update idletasks
        }} $im]
      trace add variable overlays($id,Color) write [list apply {{im v1 v2 op} {
          upvar $v1 overlays
          $im put $overlays($v2) -to 0 0 10 10
          update idletasks
        }} $im]
    }
    trace add variable overlays([lindex [array names overlays $id,*] 0]) unset [list apply {{win item v1 v2 op} {
        $win delete $item
    }} $win $item]
    installUpdFunc $id
  }
  $win selection set $item
  return $ret
}

proc ::overlay::installUpdFunc {id} {
  variable updateFunc
  variable updateProps
  variable overlays
  # install a trace for an external update function
  if {[llength $updateFunc] >= 1} {
    if {[llength $updateFunc] == 3} {  
      # last element is a delay in ms, schedule the call of the update function
      if {$updateProps eq {}} {
        set names [array names overlays $id,*]
      } else {
        foreach prop $updateProps {
          lappend names ${id},${prop}
        }
      }
      foreach name $names {
        trace add variable overlays($name) write [list apply {{updFunc updArgs delay id v1 v2 op} {
            regexp -- {^[[:digit:]]+,([[:print:]]+)$} $v2 -> prop
            ::overlay::schedUpdFunc $updFunc $updArgs $delay $id $v1 $prop $op
          }
        } {*}$updateFunc $id]
        trace add variable overlays($name) unset [list apply {{updFunc updArgs delay id v1 v2 op} {
            regexp -- {^[[:digit:]]+,([[:print:]]+)$} $v2 -> prop
            ::overlay::schedUpdFunc $updFunc $updArgs $delay $id $v1 $prop $op
          }
        } {*}$updateFunc $id]
      }
      # since the overlay is already added (and thus the traced variable set)
      # call the scheduler now
      regexp -- {^[[:digit:]]+,([[:print:]]+)$} [lindex [array names overlays $id,*] 0] -> prop
      DEBUGPRINT "Calling update function [lindex $updateFunc 0] after [lindex $updateFunc end] ms delay"
      ::overlay::schedUpdFunc {*}$updateFunc $id overlays $prop write
    } else {
      # call the update function immediately
      if {$updateProps eq {}} {
        set names [array names overlays $id,*]
      } else {
        foreach prop $updateProps {
          lappend names ${id},${prop}
        }
      }
      foreach name $names {
        trace add variable overlays($name) write [list apply {{updFunc updArgs id v1 v2 op} {
            regexp -- {^[[:digit:]]+,([[:print:]]+)$} $v2 -> prop
            $updFunc $updArgs $id $v1 $prop $op
          }
        } {*}$updateFunc $id]
        trace add variable overlays($name) unset [list apply {{updFunc updArgs id v1 v2 op} {
            regexp -- {^[[:digit:]]+,([[:print:]]+)$} $v2 -> prop
            $updFunc $updArgs $id $v1 $prop $op
          }
        } {*}$updateFunc $id]
      }
      # since the overlay is already added (and thus the traced variable set)
      # call the update function now
      regexp -- {^[[:digit:]]+,([[:print:]]+)$} [lindex [array names overlays $id,*] 0] -> prop
      DEBUGPRINT "Calling update function $updateFunc"
      eval {*}$updateFunc $id overlays $prop write
    }
  }
}

# create an event handler, store the result in a global variable named after the update function
proc ::overlay::schedUpdFunc {updFunc updArgs delay id v1 v2 op} {
  catch {after cancel [set ::overlay::$updFunc]}
  # set the variable in the ::overlay namespace
  set ::overlay::$updFunc [after $delay $updFunc $updArgs $id $v1 $v2 $op]
}

